signature XML =
sig
  type attrs = (string * string) list
  type meta  = (string * attrs) list
  type node
  type xml

  val mk_node : string -> attrs -> node list -> node
  val mk_text_node : string -> attrs -> node list -> string -> node
  (* The result of [mk_problem trs] is used as input for an
     external termination tool (after applying [to_string] to it). *)
  val mk_problem : term (* IsaFoR Q-list *) -> term option (* IsaFoR DPs *) -> term (* IsaFoR TRS *) -> xml

  val to_string : xml -> string
end

structure Xml : XML = 
struct
  type attrs = (string * string) list
  type meta  = (string * attrs) list
  datatype node  = Node of {
    id    : string,
    attrs : attrs,
    succs : node list,
    text  : string option
  }
  type xml = (meta * node)

  (* Transforming XML trees into strings *)
  infixr 3 +#+
  infixr 3 +@+

  fun s +@+ t = s o t
  fun shows_string s t = s ^ t
  fun s +#+ t = shows_string s +@+ t
  val empty = I
  val shows_nl = shows_string "\n"

  fun shows_map _ [] = I
    | shows_map show (x::xs) = show x +@+ shows_map show xs

  fun shows_attr (k,v) = k +#+ "=\"" +#+ v +#+ "\"" +#+ empty
  val shows_attrs = shows_map (fn a => " " +#+ shows_attr a)
  fun shows_tag b e id attrs = "<" +#+ b +#+ id +#+ shows_attrs attrs +@+ e +#+ ">" +#+ empty

  val shows_meta = shows_map (fn (id,attrs) => shows_tag "?" "?" id attrs +@+ shows_nl) 

  val shows_node =
    let
      fun shows i (Node{id,attrs,succs=[],text=NONE})   = i +@+ shows_tag "" "/" id attrs
	      | shows i (Node{id,attrs,succs=[],text=SOME t}) =
	          i +@+ shows_tag "" "" id attrs +@+ t +#+ shows_tag "/" "" id []
	      | shows i (Node{id,attrs,succs,text})           = 
	          i
	          +@+ shows_tag "" "" id attrs
	          +@+ shows_map (shows (i +@+ "  " +#+ empty)) succs
	          +@+ (if text=NONE then empty else shows_string (the text))
	          +@+ i
	          +@+ shows_tag "/" "" id []
    in shows shows_nl end

  fun shows_xml (meta,root) = shows_meta meta +@+ shows_node root
  fun to_string xml = shows_xml xml ""

  (* Constructing XML trees *)

  fun mk_node id attrs succs = Node{id=id,attrs=attrs,succs=succs,text=NONE}
  fun mk_text_node id attrs succs txt = Node{id=id,attrs=attrs,succs=succs,text=SOME txt}

  val mk_term = Isafor_Basics.fold_term
		 (mk_text_node "var" [] [])
		 (fn f => fn ts => mk_node "funapp" []
					   (mk_text_node "name" [] [] f
					    :: map (fn x => mk_node "arg" [] [x]) ts))
		 (fn t => raise TERM ("mk_term",[t]))

  fun mk_rule p =
    let
      val (lhs,rhs) = HOLogic.dest_prod p
    in
      mk_node "rule" [] [mk_node "lhs" [] [mk_term lhs],mk_node "rhs" [] [mk_term rhs]]
    end

  fun mk_signature trs =
    let
      fun mk_fa (f,n) = mk_node "funcsym" [] [
        mk_text_node "name" [] [] f,
       	mk_text_node "arity" [] [] (string_of_int n)
      ]
    in
      mk_node "signature" [] (map mk_fa (Isafor_Term.funas_trs trs))
    end

  fun mk_trs name trs = 
    let
      val trs = HOLogic.dest_list trs
    in 
      mk_node name [] [
        mk_node "rules" [] (map mk_rule trs),
        mk_signature trs]
    end

  fun mk_strategy q = 
    let
      val ql = HOLogic.dest_list q
    in 
      mk_node "innermost_lhss" [] 
        (map mk_term ql)       
    end

  val meta = [("xml",[("version","1.0"),("encoding","UTF-8"),("standalone","no")])]

  val msg = mk_node "metainformation" [] [
    mk_text_node "comment" [] []
      "This file has automatically been generated by Isabelle/HOL/IsaFoR"]

  fun mk_problem q dp_opt r = (meta,mk_node
    "problem"
    [("xmlns:xsi","http://www.w3.org/2001/XMLSchema-instance"),
     ("type","termination"),
     ("xsi:noNamespaceSchemaLocation","http://dev.aspsimon.org/xtc.xsd")]
    ((case dp_opt of NONE => [] | SOME dps => [mk_trs "pairs" dps]) @ [mk_trs "trs" r,mk_text_node "strategy" [] [] "FULL",mk_strategy q,msg]))
end
